'''
Created on 2011-11-18

@author: Andrew Roth
'''
from math import log, lgamma as log_gamma

#=======================================================================================================================
# Constant node classes.
#=======================================================================================================================
class ScalarConstantNode(object):
    '''
    A constant valued scalar node.
    '''
    def __init__(self, value):
        self.value = value
    
    @property
    def e_x(self):
        return self.value
    
    @property
    def e_x2(self):
        return self.value ** 2
    
    @property
    def e_log_x(self):
        return log(self.value)
    
    @property
    def e_log_gamma_x(self):
        return log_gamma(self.value)
            
class VectorConstantNode(object):
    '''
    A constant vector valued node.
    '''
    def __init__(self, value):
        '''
        value : A list of length d, where d is the dimension of the node.
        '''
        self.value = value
        
        self.dim = len(value)
    
    @property
    def e_x(self):
        return self.value
    
    @property
    def e_log_x(self):
        return [log(x) for x in self.value]

#=======================================================================================================================
# Abstract base class for nodes.
#=======================================================================================================================
class DistributionNode(object):
    '''
    Abstract base class for all distribution node types.
    '''
    def __init__(self, parents, value=None):
        self.value = value
        
        if value is not None:
            self.observed = True
        else:
            self.observed = False
        
        self.parents = {}
        self.children = set()
        
        self._set_parents(parents)
        
        self._init_ens()
    
    # Abstract methods.
    def _set_ens_from_observed_values(self):
        '''
        Use the observed values to initialise ens. Implementation will vary between the abstract scalar and vector
        nodes classes. This should only be called if the node is observed.
        '''
        raise NotImplemented
    
    def _set_ens_from_parent_values(self):
        '''
        Use the values of parent variables to initialise ens. Implementation will depend on details of the
        distribution node. This should only be called if a node is not observed.
        '''
        raise NotImplemented
        
    def _set_parents(self):
        '''
        Add parents to parent dictionary, and add this node as child to the parents if parent is not a ConstantNode.
        ''' 
        raise NotImplemented
    
    def get_parent_message(self, parent):
        '''
        Return appropriate message to parent.
        '''
        raise NotImplemented
    
    def update_prior(self):
        raise NotImplemented
    
    def update_ens(self):
        '''
        Update the expected natural statistics of the distribution. This should be done using the posterior_parameters
        of the distribution.
        '''
        raise NotImplemented    
    
    # Implemented methods
    def _init_ens(self):
        '''
        Initialise first and second ens of node.
        ''' 
        self.ens = [None] * 2
        
        if self.value is not None:
            self._set_ens_from_observed_values()
        else:
            self._set_ens_from_parent_values()    
    
    def add_child(self, child_node):
        '''
        Add a node to the set of children.
        '''
        self.children.add(child_node)
    
    def update(self):
        '''
        Update a node. This collects the messages from all children and updates the node.
        '''
        self.update_posterior()
        self.update_ens()
    
    def update_posterior(self):
        '''
        Update the posterior parameters of the node, by collecting and summing all child messages.
        '''
        # Copy posterior parameters       
        self.posterior_parameters = self.prior_parameters[:]
        
        for child in self.children:
            message = child.get_parent_message(self)
            
            for i, m in enumerate(message):
                self.posterior_parameters[i] += m

    def get_bound(self):
        '''
        Compute the contribution to the lower bound from this node.
        '''
        if self.observed:
            return self.get_expected_log_p()
        else:
            return self.get_expected_log_p() - self.get_expected_log_q()
    
    def get_expected_log_p(self):
        '''
        Get the expected value of the log of the actual distribution for this node.
        '''
        self.update_prior()
        
        expected_log_p = 0

        for prior_parameter, ens in zip(self.prior_parameters, self.ens):
            expected_log_p += prior_parameter * ens
        
        expected_log_p += self.get_prior_g()
        
        return expected_log_p                   

    def get_expected_log_q(self):
        '''
        Get the expected value of the log of the variational approximating distribution for this node.
        ''' 
        expected_log_q = 0;
        
        for posterior_parameter, ens in zip(self.posterior_parameters, self.ens):
            expected_log_q += posterior_parameter * ens
            
        expected_log_q += self.get_posterior_g()
        
        return expected_log_q
    
    def _is_node(self, obj):
        '''
        Check if object is a node. The alternative is a scalar, or list of values.
        '''
        if hasattr(obj, 'value'):
            return True
        else:
            return False
    
    @property
    def parent_child_list(self):
        '''
        Return a list of parent child tuples objects.
        '''
        return [(x, self) for x in self.parents.values()]

class ScalarDistributionNode(DistributionNode):
    '''
    Abstract base class for distributions over a single variable.
    '''
    # Implemented methods
    def _set_parents(self, parents):
        for name, value in parents.items():
            if self._is_node(value):
                self.parents[name] = value
            else:
                self.parents[name] = ScalarConstantNode(value)      
        
class VectorDistributionNode(DistributionNode):
    '''
    Abstract base class for distributions over multiple variables.
    '''    
    def _set_parents(self, parents):
        for name, value in parents.items():
            if self._is_node(value):
                self.parents[name] = value
            else:
                self.parents[name] = VectorConstantNode(value)
        
        parent_dims = []
        
        for value in self.parents.values():
            parent_dims.append(value.dim)
        
        for dim in parent_dims:
            if dim != parent_dims[0]:
                raise Exception("Dimensions of multivariate parent nodes must be the same.")
        
        self.dim = parent_dims[0]

#=======================================================================================================================
# Exceptions
#=======================================================================================================================
class InvalidParentException(Exception):
    pass
