'''
Class for representing a mixture distribution.

Created on 2011-11-18

@author: Andrew Roth
'''
from pyvmp.nodes.core import InvalidParentException, DistributionNode, VectorConstantNode

class MixtureDistribution(DistributionNode):
    def __init__(self, component_class, parents, value=None):
        self.component_class = component_class
        
        self.num_components = len(parents['components'])
        
        self.value = value
        
        if value is not None:
            self.observed = True
        else:
            self.observed = False
        
        self.parents = {}
        self.children = set()
        
        self._set_parents(parents)
        
        self._init_ens()
        
        self.prior_parameters = [float('nan')] * len(self.components[0].prior_parameters)
        
        self.update_prior()
    
    def _set_parents(self, parents):
        index = parents['index']

        if self._is_node(index):
            self.parents['index'] = index
        else:
            self.parents['index'] = VectorConstantNode(index)
        
        self.components = []
        
        self.parents['components'] = []
        
        for component_parents in parents['components']:
            component = self.component_class(component_parents, value=self.value)
                        
            self.components.append(component)
            
            self.parents['components'].append(component.parents)
    
    def _init_ens(self):
        '''
        Initialise first and second ens of node.
        '''
        self.ens = [0] * len(self.components[0].ens)
        
        if self.value is not None:
            self._set_ens_from_observed_values()
        else:
            self._set_ens_from_parent_values()   
    
    def _set_ens_from_observed_values(self):
        self.ens = self.components[0].ens[:]
        
    def _set_ens_from_parent_values(self):
        '''
        Initialise by taking a weighted sum of ens from the components.
        '''        
        for index, component in enumerate(self.components):
            for j, ens in enumerate(component.ens):
                self.ens[j] += self.mix_weights[index] * ens
    
    # Message functions
    def get_parent_message(self, parent):
        if parent is self.parents['index']:
            message = self._index_message()
        else:
            message = self._components_message(parent)
        
        return message
    
    def _index_message(self):
        return [x.get_expected_log_p() for x in self.components]

    def _components_message(self, parent):
        '''
        To compute a message to component, we get the message that would be sent normally to the parent and multiply it
        by the mix-weight for that component.
        '''
        indices = self._get_component_indices(parent)
        
        if indices is None:
            # If we get here no parent has been found
            raise InvalidParentException
        
        message = [0.] * len(self.components[0].ens)
        
        for index in indices:
            # Ignore near empty clusters.
            if self.mix_weights[index] < 1e-10:
                continue
                        
            component = self.components[index]
            
            component_message = component.get_parent_message(parent)
    
            for i, m in enumerate(component_message):
                message[i] += self.mix_weights[index] * m        
            
        return message

    # Update functions
    def update_prior(self):
        # Reset the prior parameters to 0
        self.prior_parameters = [0 for _ in self.prior_parameters]
        
        # Update the prior parameters by taking a weighted sum of parent priors
        for index, component in enumerate(self.components):
            component.update_prior()
            
            for j, component_prior in enumerate(component.prior_parameters):
                self.prior_parameters[j] += self.mix_weights[index] * component_prior
        
    def update_ens(self):
        self.ens = [0 for _ in self.ens]
        
        for component in self.components:
            component.update_ens()
    
    # Lower bound functions
    def get_prior_g(self):
        prior_g = 0
        
        for index, component in enumerate(self.components):
            prior_g += self.mix_weights[index] * component.get_prior_g()
        
        return prior_g

    def get_posterior_g(self):
        posterior_g = 0
        
        for index, component in enumerate(self.components):
            posterior_g += self.mix_weights[index] * component.get_posterior_g()
        
        return posterior_g
    
    # Helper functions
    @property
    def mix_weights(self):
        '''
        The mix-weights are just the ens of the index parent.
        '''
        return self.parents['index'].e_x
        
    @property
    def parent_child_list(self):
        parent_child = [(self.parents['index'], self)]
        
        for comp_parents in self.parents['components']:
            parent_child.extend([(x, self) for x in comp_parents.values()])
        
        return parent_child
    
    def _get_component_indices(self, parent):
        '''
        Retrieve the index of the components which have parent as a parent. If no component with such a parent is found 
        return None.
        '''
        indices = []
        
        for index, component in enumerate(self.components):
            if parent in component.parents.values():
                indices.append(index)
        
        if indices:
            return indices
        else:
            return None
            
