'''
Created on 2011-11-16

@author: Andrew Roth
'''
from __future__ import division

from math import log, lgamma as log_gamma, pi
from pyvmp.utils import psi

from pyvmp.nodes.core import ScalarDistributionNode, InvalidParentException

class Gamma(ScalarDistributionNode):
    def __init__(self, parents, value=None):
        ScalarDistributionNode.__init__(self, parents, value)

        self.prior_parameters = [float('nan')] * 2
        self.update_prior()
        
        self.posterior_parameters = self.prior_parameters[:]
        
    def _set_ens_from_observed_values(self):
        self.ens[0] = self.value
        self.ens[1] = log(self.value)
    
    def _set_ens_from_parent_values(self):
        a = self.parents['a'].e_x
        b = self.parents['b'].e_x

        self.ens[0] = a / b
        self.ens[1] = psi(a) - log(b)

    # Message functions.
    def get_parent_message(self, parent):
        if parent is self.parents['b']:
            message = self._scale_message()
        elif parent is self.parents['a']:
            raise Exception('Gamma nodes cannot currently have stochastic shape parameters.')
        else:
            raise InvalidParentException
        
        return message

    def _scale_message(self):
        return [
                - self.e_x,
                self.parents['a'].e_x
                ]

    # Update functions
    def update_prior(self):        
        self.prior_parameters[0] = -self.parents['b'].e_x
        self.prior_parameters[1] = self.parents['a'].e_x - 1
        
    def update_ens(self):
        a, b = self.get_posterior_standard_parameters()

        self.ens[0] = a / b        
        self.ens[1] = psi(a) - log(b)
            
    # Lower bound functions
    def get_prior_g(self):
        a = self.parents['a'].e_x
        
        log_gamma_a = self.parents['a'].e_log_gamma_x
        
        log_b = self.parents['b'].e_log_x
        
        return self._evaluate_g(a, log_b, log_gamma_a)


    def get_posterior_g(self):
        a, b = self.get_posterior_standard_parameters()
        
        log_b = log(b)
        
        log_gamma_a = log_gamma(a)
        
        return self._evaluate_g(a, log_b, log_gamma_a)
    
    def _evaluate_g(self, a, log_b, log_gamma_a):
        return a * log_b - log_gamma_a
    
    # Helper functions
    def get_posterior_standard_parameters(self):
        a = self.posterior_parameters[1] + 1        
        b = -self.posterior_parameters[0]
        
        return a, b
    
    @property
    def e_x(self):
        '''
        Return first moment.
        '''
        return self.ens[0]
    
    @property
    def e_log_x(self):
        '''
        Return the expected value of log(x).
        '''
        return self.ens[1]
        
class Gaussian(ScalarDistributionNode):
    '''
    Gaussian distribution.
    
    Natural Statistics : [x, x**2]
    Natural Parameters : [mean * precision, - 1/2 * precision]
    '''    
    def __init__(self, parents, value=None):
        ScalarDistributionNode.__init__(self, parents, value)
        
        self.prior_parameters = [float('nan')] * 2       
        self.update_prior()
        
        self.posterior_parameters = self.prior_parameters[:]
    
    def _set_ens_from_observed_values(self):
        self.ens[0] = self.value
        self.ens[1] = self.value ** 2
        
    def _set_ens_from_parent_values(self):
        mean = self.parents['mean'].e_x
        precision = self.parents['precision'].e_x
        
        variance = (1 / precision)
        
        self.ens[0] = mean            
        self.ens[1] = mean ** 2 + variance
    
    # Message functions
    def get_parent_message(self, parent):
        if parent is self.parents['mean']:
            message = self._mean_message()

        elif parent is self.parents['precision']:
            message = self._precision_message()
        else:
            raise InvalidParentException

        return message
    
    def _mean_message(self):
        x_1 = self.e_x
        m_1 = self.parents['precision'].e_x
        
        message = [
                   x_1 * m_1 ,
                   - 0.5 * m_1
                   ]

        return message
    
    def _precision_message(self):
        x_1 = self.e_x
        x_2 = self.e_x2
        
        m_1 = self.parents['mean'].e_x
        m_2 = self.parents['mean'].e_x2
        
        message = [
                   x_1 * m_1 - 0.5 * (x_2 + m_2),
                   0.5
                   ]
        
        return message

    # Update functions
    def update_prior(self):
        mean = self.parents['mean'].e_x
        precision = self.parents['precision'].e_x
        
        self.prior_parameters[0] = mean * precision
        self.prior_parameters[1] = -0.5 * precision

    def update_ens(self):
        mean, precision = self.get_posterior_standard_parameters()
        
        variance = 1 / precision
        
        if variance < 0: 
            raise Exception("Negative variance {0}".format(variance))

        self.ens[0] = mean
        self.ens[1] = mean ** 2 + variance    
    
    # Lower bound functions
    def get_prior_g(self):
        mean_squared = self.parents['mean'].e_x2
        
        precision = self.parents['precision'].e_x
        
        log_precision = self.parents['precision'].e_log_x
        
        return self._evaluate_g(mean_squared, precision, log_precision)

    def get_posterior_g(self):
        mean, precision = self.get_posterior_standard_parameters()
        
        mean_squared = mean ** 2
        
        log_precision = log(precision)

        return self._evaluate_g(mean_squared, precision, log_precision)

    def _evaluate_g(self, mean_squared, precision, log_precision):
        return 0.5 * log_precision - \
               0.5 * precision * mean_squared - \
               0.5 * log(2 * pi)
               
    
    def get_posterior_standard_parameters(self):        
        precision = -2 * self.posterior_parameters[1]
        
        mean = self.posterior_parameters[0] / precision
        
        return mean, precision
    
    @property
    def e_x(self):
        '''
        Return the first moment.
        '''
        return self.ens[0]
    
    @property
    def e_x2(self):
        '''
        Return the second moment.
        '''
        return self.ens[1]