'''
Created on 2011-11-18

@author: Andrew Roth
'''
from __future__ import division

from math import exp, log, lgamma as log_gamma
from pyvmp.utils import psi

from pyvmp.nodes.core import VectorDistributionNode, VectorConstantNode, InvalidParentException
from pyvmp.utils import log_space_normalise

class Discrete(VectorDistributionNode):
    def __init__(self, parents, value=None):
        VectorDistributionNode.__init__(self, parents, value)
        
        self.prior_parameters = [float('nan')] * self.dim
        self.update_prior()
        
        self.posterior_parameters = self.prior_parameters[:]
        
    def _set_ens_from_observed_values(self):        
        self.ens = [x for x in self.value]        

    def _set_ens_from_parent_values(self):
        self.ens = [x for x in self.parents['pi'].e_x]
    
    # Message functions
    def get_parent_message(self, parent):
        return self.ens[:]
    
    # Update functions
    def update_prior(self):
        self.prior_parameters = self.parents['pi'].e_log_x

    def update_ens(self):
        self.posterior_parameters = log_space_normalise(self.posterior_parameters)
        
        self.ens = [exp(x) for x in self.posterior_parameters]

    # Lower bound functions
    def get_prior_g(self):
        return 0

    def get_posterior_g(self):
        return 0
    
    @property
    def e_x(self):
        return self.ens[:]
    
class ConditionalDiscrete(VectorDistributionNode):
    def __init__(self, parents, value=None):
        VectorDistributionNode.__init__(self, parents, value)
        
        self.prior_parameters = [float('nan')] * self.dim
        self.update_prior()
        
        self.posterior_parameters = self.prior_parameters[:]
        
    def _set_parents(self, parents):
        X = parents['X']
        
        if not self._is_node(X):
            X = VectorConstantNode(X)        
        
        self.parents['X'] = X

        self.parents['A'] = []
        
        for A in parents['A']:
            if not self._is_node(A):
                A = VectorConstantNode(A)
            
            self.parents['A'].append(A)
        
        self.dim = self.parents['A'][0].dim
        self.parent_dim = self.parents['X'].dim
            
    def _set_ens_from_observed_values(self):        
        self.ens = [x for x in self.value]        

    def _set_ens_from_parent_values(self):
        self.ens = [0.] * self.dim
        
        for i in range(self.dim):
            for j in range(self.parent_dim):
                self.ens[i] = self.parents['X'].e_x[j] * self.parents['A'][j].e_x[i]
    
    def get_parent_message(self, parent):
        if parent is self.parents['X']:
            message = self._discrete_message(parent)
        elif parent in self.parents['A']:
            message = self._dirichlet_message(parent)
        else:
            raise InvalidParentException
        
        return message
        
    def _discrete_message(self, parent):
        message = [0] * parent.dim
        
        for i in range(parent.dim):
            log_A = self.parents['A'][i].e_log_x            
            
            for j in range(self.dim):
                message[i] += self.e_x[j] * log_A[j]
        
        return message
    
    def _dirichlet_message(self, parent):
        index = self.parents['A'].index(parent)
        
        message = [0.] * self.dim
        
        for i in range(self.dim):
            message[i] = self.parents['X'].e_x[index] * self.e_x[i]
        
        return message

    def update_prior(self):
        self.prior_parameters = [0.] * self.dim
        
        X = self.parents['X'].e_x
        
        for i in range(self.parent_dim):            
            log_A = self.parents['A'][i].e_log_x
            
            for j in range(self.dim):                
                self.prior_parameters[j] += log_A[j] * X[i]

    def update_ens(self):
        self.posterior_parameters = log_space_normalise(self.posterior_parameters)
        
        self.ens = [exp(x) for x in self.posterior_parameters]

    # Lower bound functions
    def get_prior_g(self):
        return 0

    def get_posterior_g(self):
        return 0
    
    @property
    def e_x(self):
        return self.ens[:]
    
    @property
    def parent_child_list(self):
        parent_child = [(self.parents['X'], self)]
        
        for A in self.parents['A']:
            parent_child.append((A, self))
        
        return parent_child

class Dirichlet(VectorDistributionNode):
    def __init__(self, parents, value=None):        
        VectorDistributionNode.__init__(self, parents, value)
        
        self.prior_parameters = [float('nan')] * self.dim
        self.update_prior()
        
        self.posterior_parameters = self.prior_parameters[:]

    def _set_ens_from_observed_values(self):        
        self.ens = [log(x) for x in self.value]
        
    def _set_ens_from_parent_values(self):
        alpha = self.parents['alpha'].e_x
        
        psi_sum = psi(sum(alpha))
              
        self.ens = [psi(x) - psi_sum for x in alpha]
    
    # Message functions
    def get_parent_message(self, parent):
        raise Exception("Dirichlet nodes do not support stochastic parents")

    def update_prior(self):
        '''
        Prior paramters are just the pseudo counts (alpha) from parent.
        '''
        alpha = self.parents['alpha'].e_x
        
        self.prior_parameters = [x - 1 for x in alpha]
        
    def update_ens(self):
        '''
        E(X_i) = alpha_i / sum(alpha)
        '''
        alpha = self.get_posterior_standard_parameters()
        
        psi_sum = psi(sum(alpha))
 
        self.ens = [psi(x) - psi_sum for x in alpha]
        
    # Lower bound functions
    def get_prior_g(self):
        alpha = self.parents['alpha'].e_x
        
        return self._evaluate_g(alpha)

    def get_posterior_g(self):
        alpha = self.get_posterior_standard_parameters()
        
        return self._evaluate_g(alpha)
    
    def _evaluate_g(self, alpha):
        return log_gamma(sum(alpha)) - sum([log_gamma(x) for x in alpha])
    
    def get_posterior_standard_parameters(self):
        return [x + 1 for x in self.posterior_parameters]
    
    @property
    def e_log_x(self):
        return self.ens[:]
    
    @property
    def e_x(self):
        alpha = self.get_posterior_standard_parameters()
        
        norm_const = sum(alpha)
        
        return [x / norm_const for x in alpha]
    
class Multinomial(VectorDistributionNode):
    def __init__(self, parents, value=None):
        VectorDistributionNode.__init__(self, parents, value)
        
        self.prior_parameters = [float('nan')] * self.dim
        
        self.update_prior()
        
        self.posterior_parameters = self.prior_parameters[:]
        
    def _set_ens_from_observed_values(self):
        self.ens = [x for x in self.value]        

    def _set_ens_from_parent_values(self):
        raise Exception("Only observed Multinomials allowed.")
    
    # Message functions
    def get_parent_message(self, parent):
        return self.ens[:]
    
    # Update functions
    def update_prior(self):
        self.prior_parameters = self.parents['pi'].e_log_x

    def update_ens(self):
        raise Exception("Only observed Multinomials allowed.")

    # Lower bound functions
    def get_prior_g(self):
        return 0

    def get_posterior_g(self):
        return 0
    
    @property
    def e_x(self):
        return self.ens[:]    
